/*
    Fichier cylcree.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Programme permettant de crer un fichier cyloop.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "types.h"
#include "limites.h"
#include "messages.h"
#include "dialogue.h"

#define  szmax_nomfic   80


// prototypes
char *recup_nomfic (char *nom);
void choix_cycle ();
void recup_info_var ();
void recap_duree ();
void gendon ();
void gendon16bits ();
void gendon32bits ();
void gendonfloat  ();



// variables globales (pour viter des tas de passages de paramtre)
FILE        *descfic;  // descripteur du fichier cyloop
desc_cyloop entetefic; // entte du fichier cyloop
desc_var    entetevar; // entte de la variable du fichier cyloop



// programme principal

main (int narg, char *varg[])
{
    char *nomfic;  // nom du fichier cyloop


    // mmorisation du nom de la commande
    memcom (*varg);

    // rcupration du nom de fichier et rajout ventuel du suffixe
    if (! --narg)
        nomfic = recup_nomfic (NULL);
    else
        nomfic = recup_nomfic (varg [1]);

    // termin si aucun nom de fichier saisi (juste l'extention .cyl)
    if (strlen (nomfic) < 5)
        return;

    // on n'crase pas un fichier cyloop qui existe dj
    if (access (nomfic, 0) == 0)
        // "fichier cyloop %s dj existant"
        err_fatale_arg ("CYLOOP_EXISTE", nomfic);

    // cration du fichier cyloop
    descfic = fopen (nomfic, "w");

    if (! descfic)
        // "Cration impossible du fichier cyloop %s"
        err_fatale_arg ("PB_CREE_CYLOOP", nomfic);

    // initialisation
    entetefic.signature [0] = 'c';  // signature du fichier cyloop
    entetefic.signature [1] = 'y';
    entetefic.signature [2] = 'l';
    entetefic.nb_var = 0;  // fichier cyloop de niveau 0 : une variable sans nom

    // choix du cycle de mesures
    choix_cycle ();

    // caractristiques de la variable mmorisant les cycles
    recup_info_var ();

    // enregistrement de l'entte du fichier cyloop
    // on ne recopie pas les ventuels octets redondants de la structure
    fwrite (&entetefic, 16, 1, descfic);

    // enregistrement de l'entte de la variable du fichier cyloop
    if (entetevar.typedon & cmoy_sep)
        // une valeur de coef_moy pour chaque instant du cycle
        fwrite (&entetevar, 6, 1, descfic);
    else
        // champ coef_moy global pour la variable
        fwrite (&entetevar, 8, 1, descfic);

    // gnration des donnes avec leur valeur initiale
    gendon ();

    // le fichier cyloop est prt
    if (fclose (descfic) < 0)
        // "Problme d'criture dans le fichier cyloop"
        affiche_err ("PB_ECR_CYLOOP");
}



// rcupration du nom du fichier cyloop et rajout ventuel de son suffixe

char *recup_nomfic (char *nom)
{
    char  *nomfic;
    int   lg_nomfic;


    // si un nom de fichier a t pass en paramtre
    if (nom)
    {
        // rcuprer sa taille
        lg_nomfic = strlen (nom);

        // si ce nom se termine par le suffixe .cyl
        if ((lg_nomfic > 4) && (strcmp (nom + lg_nomfic - 4, ".cyl") == 0))
        {
            // mmoriser le nom tel quel
            nomfic = malloc (lg_nomfic + 1);
            strcpy (nomfic, nom);
        }
        // sinon
        else
        {
            // rajouter au nom le suffixe .cyl et mmoriser le rsultat
            nomfic = malloc (lg_nomfic + 5);
            strcpy (nomfic, nom);
            strcpy (nomfic + lg_nomfic, ".cyl");
        }
    }
    // sinon
    else
    {
        // demander le nom du fichier  l'oprateur et mmoriser sa rponse
        nomfic = malloc (szmax_nomfic);

        // "Nom du fichier ? "
        affiche_msg ("DEMANDE_NOMFICH");
        fgets (nomfic, szmax_nomfic - 4, stdin);

        // supprimer le '\n' ventuel retourn par la fonction fgets
        lg_nomfic = strlen (nomfic);

        if (nomfic [lg_nomfic - 1] == '\n')
            nomfic [--lg_nomfic] = '\0';

        // rajouter au nom le suffixe .cyl si ncessaire
        if ((lg_nomfic < 4) || (strcmp (nomfic + lg_nomfic - 4, ".cyl") != 0))
            strcpy (nomfic + lg_nomfic, ".cyl");

        // vider le buffer clavier
        fflush (stdin);
    }

    // retourner le nom du fichier
    return nomfic;
}



// choix du cycle de mesures

void choix_cycle ()
{
    struct tm stm;      // structure contenant la date et l'heure
    uint32 timep;       // date et heure dans un uint32
    int h_ete;          // mmorise si heure d't en ce moment
    int choix, utcloc;  // choix de l'oprateur
    uint32 duree;       // duree d'un cycle


    // initialisation champs caractristiques et tat
    entetefic.caract_gen  = 0;
    entetefic.etat_cyloop = 0;

    // choix du cycle
    // "Choisissez un cycle de mesure\n"
    affiche_ligne ("CHOIX_CYCLEMES");

            // "Cycle annuel dbutant le 1er janvier  0 H"
            // "Cycle mensuel dbutant le 1er jour du mois  0 H"
            // "Cycle hebdomadaire dbutant le lundi  0 H"
            // "Cycle hebdomadaire dbutant le dimanche  0 H"
            // "Cycle journalier dbutant  0 H"
            // "Autre cycle"
    choix = lecchoix ("CHOIX_CYCLEMES1", "CHOIX_CYCLEMES2",
                      "CHOIX_CYCLEMES3", "CHOIX_CYCLEMES4",
                      "CHOIX_CYCLEMES5", "CHOIX_CYCLEMES6", NULL);

    if (choix < 6)
    {
        // choisir le type d'heure
                           // "utilisation de l'heure temps universel"
                           // "utilisation de l'heure locale"
        utcloc = lecchoix ("UTIL_HEURE_TU", "UTIL_HEURE_LOC", NULL);

        if (utcloc & 1)
            entetefic.caract_gen |= temps_utc;

        // rcuprer la date et l'heure courante
        time (&timep);

        // passer  0 heures + rcuprer date et heure dans structure
        if (entetefic.caract_gen & temps_utc)
        {
            // Passage  0H par calcul si heure UTC
            timep = (timep / secjour) * secjour;

            // on rcupre la structure ensuite
            gmtime_r (&timep, &stm);
        }
        else
        {
            // si heure locale on rcupre d'abord la structure
            localtime_r (&timep, &stm);

            // mmorise si on est  l'heure d't
            h_ete = stm.tm_isdst;

            // et on modifie l'heure qu'elle contient
            stm.tm_hour = 0;
            stm.tm_min = 0;
            stm.tm_sec = 0;

            // avant de recalculer l'heure du dbut de la journe
            timep = mktime (&stm);
        }

        switch (choix)
        {
            // Cycle annuel dbutant le 1er janvier  0 H
            case 1: entetefic.caract_gen |= cycle_ireg | cycle_annuel;
                    timep -= stm.tm_yday * secjour;

                    if (stm.tm_year % 4)
                        entetefic.duree_cycle = 365 * secjour;
                    else
                        entetefic.duree_cycle = 366 * secjour;

                    break;

            // Cycle mensuel dbutant le 1er jour du mois  0 H
            case 2: entetefic.caract_gen |= cycle_ireg;
                    timep -= (stm.tm_mday -1) * secjour;

                    // la dure du mois pourrait aussi tre obtenue par calcul
                    switch (stm.tm_mon + 1)
                    {
                                 // fvrier
                        case  2: if (stm.tm_year % 4)
                                     entetefic.duree_cycle = 28 * secjour;
                                 else
                                     entetefic.duree_cycle = 29 * secjour;

                                 break;

                        case  4: // mois de 30 jours
                        case  6:
                        case  9:
                        case 11: entetefic.duree_cycle = 30 * secjour;
                                 break;

                                 // mois de 31 jours (7 mois par an)
                        default: entetefic.duree_cycle = 31 * secjour;
                    }

                    break;

            // Cycle hebdomadaire dbutant le lundi  0 H
            case 3: if (stm.tm_wday)
                        timep -= (stm.tm_wday - 1) * secjour;
                    else
                        timep -= 6 * secjour;

                    entetefic.duree_cycle = 7 * secjour;
                    break;

            // Cycle hebdomadaire dbutant le dimanche  0 H
            case 4: timep -= stm.tm_wday * secjour;
                    entetefic.duree_cycle = 7 * secjour;
                    break;

            // Cycle journalier dbutant  0 H
            case 5: entetefic.duree_cycle = secjour;
        }

        entetefic.deb_cycle = timep;

        if (!(entetefic.caract_gen & temps_utc) && h_ete)
            entetefic.etat_cyloop |= heure_ete;
    }
    else
    {
        // choix de la dure du cycle
                // "Cycle de ... jours"
                // "Cycle de ... heures"
                // "Cycle de ... minutes"
                // "Cycle de ... secondes"
        choix = lecchoix ("CYCLE_EN_JOURS", "CYCLE_EN_HEURES",
                          "CYCLE_EN_MIN", "CYCLE_EN_SEC", NULL);

        // "Dure du cycle ? "
        affiche_msg ("DEM_DUREE_CYCLE");
        duree = lirentier ();

        switch (choix)
        {
            case 1: duree = duree * 24;
            case 2: duree = duree * 60;
            case 3: duree = duree * 60;
            case 4: entetefic.duree_cycle = duree;
        }

        // choix de l'instant de dbut du cycle
        // "Dbut du cycle\n"
        affiche_ligne ("DEM_DEBUT_CYCLE");

                // "Il y a ... jours"
                // "Aujourd'hui  0 heures"
                // "Il y a ... heures"
                // "Au dbut de l'heure"
                // "Il y a ... minutes"
                // "Au dbut de la minute"
                // "Maintenant"
        choix = lecchoix ("DEM_DEB_CYCLE1", "DEM_DEB_CYCLE2",
                          "DEM_DEB_CYCLE3", "DEM_DEB_CYCLE4",
                          "DEM_DEB_CYCLE5", "DEM_DEB_CYCLE6",
                          "DEM_DEB_CYCLE7", NULL);

        if (choix == 1 || choix == 3 || choix == 5)
        {
            // "Depuis combien "
            affiche_msg ("DEPUIS_COMBIEN");

            switch (choix)
            {
                         // "de jours ? "
                case 1 : affiche_msg ("COMBIEN_JOURS");
                         break;

                         // "d'heures ? "
                case 3 : affiche_msg ("COMBIEN_HEURES");
                         break;

                         // "de minutes ? "
                case 5 : affiche_msg ("COMBIEN_MINUTES");
            }

            duree = lirentier ();
        }
        else
            duree = 0;

        // si dmarrage cal sur un dbut de journe
        if (choix <= 2)
        {
            // choisir le type d'heure
                     // "utilisation de l'heure temps universel"
                     // "utilisation de l'heure locale"
            utcloc = lecchoix ("UTIL_HEURE_TU", "UTIL_HEURE_LOC", NULL);

            if (utcloc & 1)
                entetefic.caract_gen |= temps_utc;
        }
        else
            utcloc = 1;  // pas de distinction entre heure d't et
                          // d'hiver pour les cycles de moins d'une journe

        // rcuprer la date et l'heure courante
        time (&timep);

        // calculer l'innstant de dbut du cycle
        switch (choix)
        {
            case 1: // au dbut de la journe ou il y a duree jours
            case 2: if (utcloc)
                        // simple calcul si heure UTC
                        entetefic.deb_cycle =
                                        ((timep / secjour) - duree) * secjour;
                    else
                    {
                        // sinon, on soustraira l'hure locale
                        localtime_r (&timep, &stm);
                        timep = ((timep / 3600) - stm.tm_hour) * 3600;

                        // avant de remonter de duree jours
                        entetefic.deb_cycle = timep - (duree * secjour);

                        // mmorise galement si on est  l'heure d't
                        if (stm.tm_isdst)
                            entetefic.etat_cyloop |= heure_ete;
                    }
                    break;

                    // au dbut de l'heure
            case 3: entetefic.deb_cycle = (timep / 3600) * 3600;
                    break;

                    // il y a duree heures
            case 4: entetefic.deb_cycle = ((timep / 3600) - duree) * 3600;
                    break;

                    // au dbut de la minute
            case 5: entetefic.deb_cycle = (timep / 60) * 60;
                    break;

                    // il y a duree minutes
            case 6: entetefic.deb_cycle = ((timep / 60) - duree) * 60;
                    break;

                    // maintenant
            case 7: entetefic.deb_cycle = timep;

        }
    }

    // rcapitulatif dure d'un cycle
    recap_duree ();

    // Saisie du nombre de donnes par cycle de mesure
    do
    {
        // "\nNombre de donnes par cycle de mesure ? "
        affiche_msg ("DEM_NB_DON_CYCLE");
        entetefic.nb_don_cycle = lirentier ();

        if (!entetefic.nb_don_cycle)
            // "Erreur de saisie, recommencez"
            affiche_err ("ERREUR_SAISIE");
    }
    while (!entetefic.nb_don_cycle);
}



// rcapitulatif dure d'un cycle

void recap_duree ()
{
    int jour, heure, minute; // dtail dure d'un cycle


    // "\nUn cycle de mesures "
    affiche_msg ("UN_CYCLE_MESURE");

    if (entetefic.caract_gen & cycle_ireg)
    {
        // "peut durer jusqu' "
        affiche_msg ("PEUT_DURER");

        if (entetefic.caract_gen & cycle_annuel)
            // "366 jours"
            affiche_msg ("366_JOURS");
        else
            // "31 jours"
            affiche_msg ("31_JOURS");
    }
    else
    {
        // "dure"
        affiche_msg ("CYCLE_DURE");

        jour = entetefic.duree_cycle / secjour;
        heure = (entetefic.duree_cycle % secjour) / 3600;
        minute = (entetefic.duree_cycle % 3600) / 60;

        if (jour)
        {
            printf (" %d ", jour);

            if (jour > 1)
                // "jours"
                affiche_msg ("JOURS");
            else
                // "jour"
                affiche_msg ("JOUR");
        }

        if (heure)
        {
            printf (" %d ", heure);

            if (heure > 1)
                // "heures"
                affiche_msg ("HEURES");
            else
                // "heure"
                affiche_msg ("HEURE");
        }

        if (minute)
        {
            printf (" %d ", minute);

            if (minute > 1)
                // "minutes"
                affiche_msg ("MINUTES");
            else
                // "minute"
                affiche_msg ("MINUTE");
        }
    }

    putchar ('\n');
}



// caractristiques de la variable mmorisant les cycles

void recup_info_var ()
{
    int choix;
    int valpond;


    // initialisations
    entetevar.der_enreg = 0;
    entetevar.typedon   = 0;
    entetevar.extention = 0;

    // "\nNature des donnes mmorises\n"
    // "Compteur de passages", "Valeurs"
    affiche_ligne ("NATURE_DONNEES");
    choix = lecchoix ("COMPTEUR_PASSAGE", "MEMO_VALEURS", NULL);

    if (choix == 1)
        entetevar.typedon |= compteur;

    // "\nVoulez vous faire un calcul de pondration"
    // " chaque changement de cycle ? "
    affiche_ligne ("DEM_CALC_PONDER");
    affiche_msg ("DEM_CALC_PONDER2");

    if (lirep () != 'n')
    {
        // "\nCoefficient de pondration ? "
        affiche_msg ("SAIS_COEF_PONDER");
        valpond = liponder ();

        while (valpond < 0 || valpond > 256)
        {
            // "Coefficient de pondration incorrect"
            // "Valeurs autorises de 0,0039  0,996"
            // "ou de 0,39 %  99,61 %"
            // "ou de 1/256  255/256"
            affiche_err ("ERREUR_COEF_POND");
            affiche_err ("VAL_PONDER_OK1");
            affiche_err ("VAL_PONDER_OK2");
            affiche_err ("VAL_PONDER_OK3");

            valpond = liponder ();
        }

        // la pondration minimale est de 1/256
        if (! valpond)
            valpond = 1;

        entetevar.ponderation = valpond;
    }
    else
        entetevar.ponderation = 0;

    if (entetevar.ponderation)
    {
        // "\nDoit on faire un calcul de pondration pour les cycles vides ? "
        affiche_msg ("PONDER_CYCLE_VIDE");

        if (lirep () != 'n')
            entetevar.typedon |= cpt_cycle_vide;
    }

    // "\nPour les calculs de moyenne, voulez-vous\n"
    affiche_ligne ("CALCULS_MOYENNE");

    if (entetevar.typedon & compteur)
        // "un coefficient global (conseill)"
        // "un coefficient distinct pour chaque instant du cycle"
        choix = lecchoix ("COEF_GLOBAL_CONS", "COEF_DISTINCT", NULL);
    else
        // "un coefficient global pour tout le cycle"
        // "un coefficient distinct pour chaque instant du cycle"
        choix = lecchoix ("COEF_GLOBAL", "COEF_DISTINCT", NULL);

    if (choix == 1)
        entetevar.coef_moy = 0;
    else
        entetevar.typedon |= cmoy_sep;

    // "\nDurant un cycle, chaque instant peut il tre pris en compte\n"
            // "une fois", "plusieurs fois"
    affiche_ligne ("CPT_INSTANTS");
    choix = lecchoix ("UNE_FOIS", "PLUSIEURS_FOIS", NULL);

    if (choix == 2)
        entetevar.typedon |= cumul_donnees;

    if ((entetevar.typedon & cmoy_sep) &&
        (entetevar.typedon & cumul_donnees))
    {
        // "\nLe coefficient de calcul de la moyenne doit il tre mis  jour\n"
                // "une fois pour chaque instant du cycle"
                // "autant de fois que l'est la variable mmorise"
        affiche_ligne ("MAJ_COEF_MOY");
        choix = lecchoix ("UNE_FOIS_INSTANT", "AUTANT_FOIS_VAR", NULL);

        if (choix == 2)
            entetevar.typedon |= cpt_toute_don;
    }

    // "\nFormat des donnes mmorises\n"
    affiche_ligne ("DEM_FORM_DONNEES");

    // si la variable est un compteur
    if (entetevar.typedon & compteur)
    {
        // choix de la taille des donnes mmorises
        if (entetevar.typedon & cumul_donnees)
            // "entier 16 bits", "entier 32 bits"
            choix = lecchoix ("ENTIER_16_BITS", "ENTIER_32_BITS", NULL);
        else
            // "entier 16 bits (suffisant)", "entier 32 bits"
            choix = lecchoix ("ENTIER_16_B_SUF", "ENTIER_32_BITS", NULL);

        if (choix == 2)
            entetevar.typedon |= donnees_32bits;
    }
    // sinon la variable mmorise une valeur  chaque instant des cycles
    else
    {
        // choix de la nature des donnes mmorises
        // "entier 32 bits", "nombre flottant"
        choix = lecchoix ("ENTIER_32_BITS", "NOMBRE_FLOTTANT", NULL);

        entetevar.typedon |= donnees_32bits;

        if (choix == 2)
            entetevar.typedon |= val_float;

        // mmorisation des valeurs min et max des donnes ?
        // "\nVoulez vous mmoriser les valeurs"
        // "min et max de chaque instant du cycle ? "
        affiche_ligne ("MEM_MIN_MAX_1");
        affiche_msg ("MEM_MIN_MAX_2");

        if (lirep () != 'n')
            entetevar.typedon |= mem_min_max;
    }
}



// gnration des donnes avec leur valeur initiale

void gendon ()
{
    // on appele une fonction dpendant du format des donnes

    if (entetevar.typedon & compteur)
    {
        if (entetevar.typedon & donnees_32bits)
            gendon32bits ();
        else
            gendon16bits ();
    }
    else
    {
        if (entetevar.typedon & val_float)
            gendonfloat ();
        else
            gendon32bits ();
    }
}



// gnration des donnes d'une variable 16 bits avec leur valeur initiale

void gendon16bits ()
{
    // la structure de donne maximale pour une variable 16 bits
    struct
    {
        uint16 coef_moy;
        int16  total;
    } donnees;

    int nb_octets; // nombre d'octets utiliss dans la structure ci-dessus
    void *debdon;  // adresse de dbut des donnes recopies dans le fichier
    int i;         // compteur de boucle


    // initialiser les champs de la structure
    // et calculer debdon et nb_octets

    if (entetevar.typedon & cmoy_sep)
    {
        donnees.coef_moy = 0;
        debdon = &donnees;
        nb_octets = 4;
    }
    else
    {
        debdon = &donnees.total;
        nb_octets = 2;
    }

    donnees.total = 0;

    // recopier les donnes dans le fichier cyloop
    for (i = 0; i < entetefic.nb_don_cycle; i++)
        if (! fwrite (debdon, nb_octets, 1, descfic))
            // "Problme d'criture dans le fichier cyloop"
            err_fatale ("PB_ECR_CYLOOP");
}



// gnration des donnes d'une variable 32 bits avec leur valeur initiale

void gendon32bits ()
{
    // la structure de donne maximale pour une variable 32 bits
    struct
    {
        uint16 vide;   // sinon coef_moy serait au mauvais endroit
        uint16 coef_moy;
        int32  total;
        int32  min;
        int32  max;
    } donnees;

    int nb_octets; // nombre d'octets utiliss dans la structure ci-dessus
    void *debdon;  // adresse de dbut des donnes recopies dans le fichier
    int i;         // compteur de boucle


    // initialiser les champs de la structure
    // et calculer debdon et nb_octets

    if (entetevar.typedon & cmoy_sep)
    {
        donnees.coef_moy = 0;
        debdon = &donnees.coef_moy;
        nb_octets = 6;
    }
    else
    {
        debdon = &donnees.total;
        nb_octets = 4;
    }

    donnees.total = 0;

    if (entetevar.typedon & mem_min_max)
    {
        donnees.min = max_int32;
        donnees.max = min_int32;
        nb_octets = nb_octets + 8;
    }

    // recopier les donnes dans le fichier cyloop
    for (i = 0; i < entetefic.nb_don_cycle; i++)
        if (! fwrite (debdon, nb_octets, 1, descfic))
            // "Problme d'criture dans le fichier cyloop"
            err_fatale ("PB_ECR_CYLOOP");
}



// gnration des donnes d'une variable non entire avec leur valeur initiale

void gendonfloat ()
{
    // la structure de donne maximale pour une variable 32 bits
    struct
    {
        uint16 vide;   // sinon coef_moy serait au mauvais endroit
        uint16 coef_moy;
        float  total;
        float  min;
        float  max;
    } donnees;

    int nb_octets; // nombre d'octets utiliss dans la structure ci-dessus
    void *debdon;  // adresse de dbut des donnes recopies dans le fichier
    int i;         // compteur de boucle


    // initialiser les champs de la structure
    // et calculer debdon et nb_octets

    if (entetevar.typedon & cmoy_sep)
    {
        donnees.coef_moy = 0;
        debdon = &donnees.coef_moy;
        nb_octets = 6;
    }
    else
    {
        debdon = &donnees.total;
        nb_octets = 4;
    }

    donnees.total = 0;

    if (entetevar.typedon & mem_min_max)
    {
        donnees.min = max_float;
        donnees.max = min_float;
        nb_octets = nb_octets + 8;
    }

    // recopier les donnes dans le fichier cyloop
    for (i = 0; i < entetefic.nb_don_cycle; i++)
        if (! fwrite (debdon, nb_octets, 1, descfic))
            // "Problme d'criture dans le fichier cyloop"
            err_fatale ("PB_ECR_CYLOOP");
}
